import Metal
import MetalKit

extension MTLComputeCommandEncoder {
	func setBytes(_ arr: [UInt32], index: Int) {
		arr.withUnsafeBytes { setBytes($0.baseAddress!, length: $0.count, index: index) }
	}
}

extension MTLRenderCommandEncoder {
	func setVertexBytes(_ arr: [UInt32], index: Int) {
		arr.withUnsafeBytes { setVertexBytes($0.baseAddress!, length: $0.count, index: index) }
	}
	func setFragmentBytes(_ arr: [UInt32], index: Int) {
		arr.withUnsafeBytes { setFragmentBytes($0.baseAddress!, length: $0.count, index: index) }
	}
}

extension MTLVertexAttributeDescriptor {
	convenience init(buf: Int, fmt: MTLVertexFormat, off: Int) {
		self.init()
		self.bufferIndex = buf
		self.format = fmt
		self.offset = off
	}
}

extension MTLVertexBufferLayoutDescriptor {
	convenience init(fn: MTLVertexStepFunction, stride: Int, rate: Int = 1) {
		self.init()
		self.stepFunction = fn
		self.stepRate = rate
		self.stride = stride
	}
}

extension MTLRenderPassColorAttachmentDescriptor {
	func setAll(_ tex: MTLTexture, load: MTLLoadAction = .load, store: MTLStoreAction = .store, color: MTLClearColor = MTLClearColor(red: 0, green: 0, blue: 0, alpha: 0)) {
		self.texture = tex
		self.loadAction = load
		self.storeAction = store
		if load == .clear {
			self.clearColor = color
		}
	}
}

extension MTLRenderPassDepthAttachmentDescriptor {
	func setAll(_ tex: MTLTexture, load: MTLLoadAction = .load, store: MTLStoreAction = .store, value: Double = 0) {
		self.texture = tex
		self.loadAction = load
		self.storeAction = store
		if load == .clear {
			self.clearDepth = value
		}
	}
}

extension MTLRenderPassStencilAttachmentDescriptor {
	func setAll(_ tex: MTLTexture, load: MTLLoadAction = .load, store: MTLStoreAction = .store, value: UInt32 = 0) {
		self.texture = tex
		self.loadAction = load
		self.storeAction = store
		if load == .clear {
			self.clearStencil = value
		}
	}
}

extension MTLDepthStencilDescriptor {
	convenience init(dcmp: MTLCompareFunction, dwrite: Bool) {
		self.init()
		self.depthCompareFunction = dcmp
		self.isDepthWriteEnabled = dwrite
	}

	convenience init(dcmp: MTLCompareFunction, dwrite: Bool, stencil: MTLStencilDescriptor) {
		self.init(dcmp: dcmp, dwrite: dwrite)
		self.frontFaceStencil = stencil
		self.backFaceStencil = stencil
	}
}

extension MTLStencilDescriptor {
	convenience init(cmp: MTLCompareFunction, pass: MTLStencilOperation, fail: MTLStencilOperation, rmask: UInt32, wmask: UInt32) {
		self.init()
		self.stencilCompareFunction = cmp
		self.depthFailureOperation = fail
		self.stencilFailureOperation = fail
		self.depthStencilPassOperation = pass
		self.readMask = rmask
		self.writeMask = wmask
	}
}

extension MTLSamplerDescriptor {
	convenience init(min: MTLSamplerMinMagFilter = .nearest, mag: MTLSamplerMinMagFilter = .nearest, mip: MTLSamplerMipFilter = .notMipmapped, normalized: Bool = true, s: MTLSamplerAddressMode = .clampToEdge, t: MTLSamplerAddressMode = .clampToEdge, r: MTLSamplerAddressMode = .clampToEdge, minLOD: Float = 0, maxLOD: Float = .greatestFiniteMagnitude) {
		self.init()
		self.minFilter = min
		self.magFilter = mag
		self.mipFilter = mip
		self.normalizedCoordinates = normalized
		self.sAddressMode = s
		self.tAddressMode = t
		self.rAddressMode = r
		self.lodMinClamp = minLOD
		self.lodMaxClamp = maxLOD
	}

	convenience init(_ fn: (MTLSamplerDescriptor) -> ()) {
		self.init()
		fn(self)
	}

	func setAddressMode(_ str: MTLSamplerAddressMode) {
		sAddressMode = str
		tAddressMode = str
		rAddressMode = str
	}

	func setClampToBorder(color: MTLSamplerBorderColor) {
		setAddressMode(.clampToBorderColor)
		borderColor = color
	}

	func setFilter(_ minMag: MTLSamplerMinMagFilter, mip: MTLSamplerMipFilter = .notMipmapped) {
		minFilter = minMag
		magFilter = minMag
		mipFilter = mip
	}

	func setLODClamp(_ clamp: ClosedRange<Float>) {
		lodMinClamp = clamp.lowerBound
		lodMaxClamp = clamp.upperBound
	}
}

func makePipeDesc(_ lib: MTLLibrary, pass: Int) -> MTLRenderPipelineDescriptor {
	let desc = MTLRenderPipelineDescriptor()
	desc.vertexFunction = lib.makeFunction(name: "vs\(pass)")!
	let vdesc = desc.vertexDescriptor!
	vdesc.layouts[30] = .init(fn: .perVertex, stride: 24)
	let attrs = vdesc.attributes
	attrs[1] = .init(buf: 30, fmt: .short4Normalized, off: 0)
	attrs[2] = .init(buf: 30, fmt: .char4Normalized, off: 8)
	attrs[4] = .init(buf: 30, fmt: .uchar4Normalized, off: 16)
	attrs[5] = .init(buf: 30, fmt: .half2, off: 20)
	if pass == 1 {
		attrs[3] = .init(buf: 30, fmt: .char4Normalized, off: 12)
		desc.colorAttachments[0].pixelFormat = .rgba8Unorm
		desc.fragmentFunction = lib.makeFunction(name: "fs_solid")!
	}
	desc.depthAttachmentPixelFormat = .depth32Float
	return desc
}

for gpu in MTLCopyAllDevices() {
	print(gpu.name)

	let q = gpu.makeCommandQueue()!
	let lib = gpu.makeDefaultLibrary()!
	let pipe0 = try gpu.makeRenderPipelineState(descriptor: makePipeDesc(lib, pass: 0))
	let pipe1 = try gpu.makeRenderPipelineState(descriptor: makePipeDesc(lib, pass: 1))
	let cpipe = try gpu.makeComputePipelineState(function: lib.makeFunction(name: "test")!)
	let atomic = gpu.makeBuffer(length: 4)!
	atomic.contents().storeBytes(of: 0, as: UInt32.self)
	let ibuf = buf_index.withUnsafeBytes { gpu.makeBuffer(bytes: $0.baseAddress!, length: $0.count)! }
	let vbuf = buf_vertex.withUnsafeBytes { gpu.makeBuffer(bytes: $0.baseAddress!, length: $0.count)! }
	let tbuf = buf_tb4.withUnsafeBytes { gpu.makeBuffer(bytes: $0.baseAddress!, length: $0.count)! }
	ibuf.label = "Indices"
	vbuf.label = "Vertices"
	let tb = gpu.makeTexture(descriptor: .textureBufferDescriptor(with: .r32Uint, width: 1617408, usage: .shaderRead))!
	let tdesc = MTLTextureDescriptor.texture2DDescriptor(pixelFormat: .rgba8Unorm, width: 1440, height: 900, mipmapped: false)
	tdesc.usage = [.shaderRead, .renderTarget]
	tdesc.storageMode = .private
	let rt0 = gpu.makeTexture(descriptor: tdesc)!
	tdesc.pixelFormat = .depth32Float
	let rtd = gpu.makeTexture(descriptor: tdesc)!
	let dss0 = gpu.makeDepthStencilState(descriptor: .init(dcmp: .greater, dwrite: true))!
	let dss1 = gpu.makeDepthStencilState(descriptor: .init(dcmp: .equal, dwrite: true))!

	do {
//		let cap = MTLCaptureDescriptor()
//		cap.captureObject = q
//		cap.destination = .developerTools
//		try? MTLCaptureManager.shared().startCapture(with: cap)
		let cb = q.makeCommandBuffer()!
		do {
			let blit = cb.makeBlitCommandEncoder()!
			blit.label = "Init"
			blit.copy(
				from: tbuf,
				sourceOffset: 0,
				sourceBytesPerRow: tbuf.length,
				sourceBytesPerImage: tbuf.length,
				sourceSize: MTLSize(width: buf_tb4.count, height: 1, depth: 1),
				to: tb,
				destinationSlice: 0,
				destinationLevel: 0,
				destinationOrigin: MTLOrigin(x: 0xA2000, y: 0, z: 0)
			)
			blit.endEncoding()
		}
		do {
			let rpdesc = MTLRenderPassDescriptor()
			rpdesc.depthAttachment.setAll(rtd, load: .clear, store: .store, value: 0)
			let enc = cb.makeRenderCommandEncoder(descriptor: rpdesc)!
			enc.label = "Depth Pass"
			enc.setViewport(MTLViewport(originX: 0, originY: 900, width: 1440, height: -900, znear: 0, zfar: 1))
			enc.setDepthStencilState(dss0)
			enc.setRenderPipelineState(pipe0)
			enc.setVertexBytes(buf_vs0, index: 0)
			enc.setVertexBytes(buf_vs1, index: 1)
			enc.setVertexBytes(buf_vs4, index: 4)
			enc.setVertexBuffer(vbuf, offset: 0, index: 30)
			enc.setVertexTexture(tb, index: 60)
			enc.drawIndexedPrimitives(
				type: .triangle,
				indexCount: 300,
				indexType: .uint16,
				indexBuffer: ibuf,
				indexBufferOffset: 0,
				instanceCount: 63,
				baseVertex: 0,
				baseInstance: 0
			)
			enc.endEncoding()
		}
		do {
			let rpdesc = MTLRenderPassDescriptor()
			rpdesc.colorAttachments[0].setAll(rt0, load: .clear, store: .store, color: .init(red: 0, green: 0, blue: 0, alpha: 0))
			rpdesc.depthAttachment.setAll(rtd, load: .load, store: .store)

			let enc = cb.makeRenderCommandEncoder(descriptor: rpdesc)!
			enc.label = "Color Pass"
			enc.setViewport(MTLViewport(originX: 0, originY: 900, width: 1440, height: -900, znear: 0, zfar: 1))
			enc.setDepthStencilState(dss1)
			enc.setRenderPipelineState(pipe1)
			enc.setVertexBytes(buf_vs0, index: 0)
			enc.setVertexBytes(buf_vs1, index: 1)
			enc.setVertexBytes(buf_vs4, index: 4)
			enc.setVertexBuffer(vbuf, offset: 0, index: 30)
			enc.setVertexTexture(tb, index: 60)
			enc.drawIndexedPrimitives(
				type: .triangle,
				indexCount: 300,
				indexType: .uint16,
				indexBuffer: ibuf,
				indexBufferOffset: 0,
				instanceCount: 63,
				baseVertex: 0,
				baseInstance: 0
			)
			enc.endEncoding()
		}
		do {
			let enc = cb.makeComputeCommandEncoder()!
			enc.label = "Verify"
			enc.setComputePipelineState(cpipe)
			enc.setBuffer(atomic, offset: 0, index: 0)
			enc.setTexture(rt0, index: 0)
			enc.setTexture(rtd, index: 1)
			enc.dispatchThreadgroups(
				MTLSize(width: 1440/16, height: 900/4, depth: 1),
				threadsPerThreadgroup: MTLSize(width: 16, height: 4, depth: 1))
			enc.endEncoding()
		}
		cb.commit()
		cb.waitUntilCompleted()
		print(atomic.contents().load(as: UInt32.self))
//		MTLCaptureManager.shared().stopCapture()
	}
}
